%{
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "sglib.h"
/* Just for being able to show the line number were the error occurs.*/
extern int line;

#define PARSER
#include "expr.h"
#undef PARSER
/* Functions used during parsing (forward declarations) */
int addvar(char *,ParType);
ParType typeDefinition(ParType, ParType);
int lookup(char *);
ParType lookup_type(char *);

/* The Symbol Table*/
ST_TABLE symbolTable;
%}
/* Output informative error messages (bison Option) */
%error-verbose

/* Declaring the possible types of Symbols*/
%union{
   int i_val;
   double f_val;
   char * id;
   ParType tokentype;
   struct {
	int i_value;
	ParType type; } se;
  }
/* Token declarations and their respective types */
%token <i_val> T_num
%token <f_val> T_real
%token '('
%token ')'

%token <tokentype> T_type
%token ':'
%token ';'
%token T_V_DECL "var"
%token T_begin "begin"
%token T_end "end"
%token <id> T_id

%left '+'
%left '*'
%left ','
/* The type of non-terminal symbols*/
%type<se> expr
%type<se> term
%type<se> factor


%%
program: "begin" {symbolTable=NULL;} variable_declarations expressions "end"
	;
	
variable_declarations: /* empty */
	| var_decl variable_declarations
	;

var_decl: "var" T_id ':'  T_type ';'
              {if (!addvar($2, $4)) 
					{
					printf("Variable:: %s on line %d. ",$2,line);
					yyerror("Var already defined");}
					}
	;

expressions: /* empty*/
	| expr ';' {printf("The type of expression is: %d \n",$1.type);} expressions
	;

expr: term  { $$.type = $1.type; }
	| expr '+' term {$$.type = typeDefinition($1.type, $3.type);}
	
	;

term: factor { $$.type = $1.type; }
	| term '*' factor { $$.type = typeDefinition($1.type, $3.type);}
	;

factor: T_num {$$.type = type_integer;}
	| T_real {$$.type = type_real;}
	| T_id {if (lookup($1)) {$$.type = lookup_type($1);}
			else {printf("Variable %s NOT declared, on line %d.",$1,line);
				  yyerror("Variable Declation fault"); $1 = type_error;}
			}
	| '(' expr ')' {$$.type = $2.type;}
	;

%%
/* Function Definitions for Syntax and Semantic Analysis */
/* Type inferece regarding arithmetic expressions */
ParType typeDefinition(ParType Arg1, ParType Arg2)
{
	if (Arg1 == type_integer)
		{if (Arg2 == type_integer) {return type_integer;}
		 else if (Arg2 == type_real) {return type_real;}
		 }	 
	else if (Arg1 == type_real) 
		{if (Arg2 == type_integer || Arg2 == type_real) {return type_real;}
		}
	else {yyerror("Type missmatch"); return type_error;}
}
/* Adding a Variable entry to the ymbol table. */
int addvar(char *VariableName,ParType TypeDecl)
{    
	ST_ENTRY *newVar;
	
	if (!lookup(VariableName)) 
		{
		newVar = malloc(sizeof(ST_ENTRY));
		newVar->varname = VariableName;
		newVar->vartype = TypeDecl;
		SGLIB_LIST_ADD(ST_ENTRY, symbolTable, newVar, next_st_var);
		return 1;
		}
	else return 0; /* error */
}

/* Looking up a symbol in the symbol table. Returns 0 if symbol was not found. */

int lookup(char *VariableName){
	ST_ENTRY *var, *result;
	var = malloc(sizeof(ST_ENTRY));
	var->varname = strdup(VariableName);
	SGLIB_LIST_FIND_MEMBER(ST_ENTRY,symbolTable,var,ST_COMPARATOR,next_st_var, result); 
	free(var);
   if (result == NULL) {return 0;}
   else {return 1;} 
}

/* Looking up a symbol type in the symbol table. Returns 0 if symbol was not found. */

ParType lookup_type(char *VariableName)
{
	ST_ENTRY *var, *result;
	var = malloc(sizeof(ST_ENTRY));
	var->varname = strdup(VariableName);
	SGLIB_LIST_FIND_MEMBER(ST_ENTRY,symbolTable,var,ST_COMPARATOR,next_st_var, result); 
	free(var);
   if (result == NULL) {return 0;}
   else {return result->vartype;} 
	
}
/* Printing the complete Symbol Table */
void print_symbol_table(void)
{
  ST_ENTRY *var;
  printf("\n Symbol Table Generated \n");
  SGLIB_LIST_MAP_ON_ELEMENTS(ST_ENTRY, symbolTable, var, next_st_var, {
    printf("ST:: Name %s Type %d \n", var->varname,var->vartype);
  });
}  
/* end of function declarations */

/* The usual yyerror */
int yyerror (const char * msg)
{
  fprintf(stderr, "PARSE ERROR: %s.\n", msg);
}

/* Main */
int main ()
{ int parse_result;
  parse_result = yyparse();

  print_symbol_table();
  
  return parse_result;
}




